/*
******************************************************************************
**  CarMaker - Version 6.* - 8.*
**  Vehicle Dynamics Simulation Toolkit
**
**  Copyright (C)   IPG Automotive GmbH
**                  Bannwaldallee 60             Phone  +49.721.98520.0
**                  76185 Karlsruhe              Fax    +49.721.98520.99
**                  Germany                      WWW    www.ipg-automotive.com
******************************************************************************
**
** Generic model extension to add traffic light detection to select TSSensors.
**
** Reference this header file in the User.c and call the register function in 
** User_Register()
**
**    TLSensor_Register ()
**
** In "Vehicle Data Set > Misc. > Additional Parameters" activate the model and
** define the number of tracked traffic lights for the desired TSSensors, e.g.
** to track two lights for a given sensor "TS01":
**
**    GenericPlugin.TLSensor.AbsPlace = DM
**    TLSensor.TS01.TrackLimit = 2
**
******************************************************************************
**
**  This example is not part of the CarMaker product. As such, it does not
**  receive official support or updates.
**
******************************************************************************
*/

#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "CarMaker.h"
#include "Car/Vehicle_Car.h"
#include "Vehicle/Sensor_TSign.h"
#include "roadbuilder.h"
#include "road.h"

#include "TLSensor.h"


//static const char ThisModelClass[] = "GenericPlugin";
static const char ThisModelKind[]  = "TLSensor";
static const int  ThisVersionId    = 1;

//struct tTLSensor 	*TLSensor;
int 				TLSensorCount = 0;


/*********************************************************************/
static void
TLSensor_DeclQuants_dyn (struct tTLSensor *mp, int park)
{
	int i, ii;
    static struct tTLSensor TLSensor_Dummy = {{0}};
    if (park)
		mp = &TLSensor_Dummy;

    /* Define here dict entries for dynamically allocated variables. */

    tDDefault *pre=DDefaultCreate("Sensor.TLight.");
	char sbuf[99];
	for (i=0; i < TLSensorCount; i++) {
		
		for (ii=0; ii < (mp[i]).trackLimit; ii++) {
			/* char sbuf[32]; */
			
			sprintf (sbuf, "%s.%d.ObjID", (mp[i]).name, ii);
			DDefInt (pre, sbuf, "", &(((mp[i]).TLight[ii]).id), DVA_None);
				
			sprintf (sbuf, "%s.%d.ds", (mp[i]).name, ii);
			DDefDouble4 (pre, sbuf, "m", &(((mp[i]).TLight[ii]).ds_p), DVA_None);
			sprintf (sbuf, "%s.%d.alpha", (mp[i]).name, ii);
			DDefDouble4 (pre, sbuf, "rad", &(((mp[i]).TLight[ii]).alpha), DVA_None);
			sprintf (sbuf, "%s.%d.theta", (mp[i]).name, ii);
			DDefDouble4 (pre, sbuf, "rad", &(((mp[i]).TLight[ii]).theta), DVA_None);
			
			sprintf (sbuf, "%s.%d.ds.x", (mp[i]).name, ii);
			DDefDouble4 (pre, sbuf, "m", &(((mp[i]).TLight[ii]).ds[0]), DVA_None);
			sprintf (sbuf, "%s.%d.ds.y", (mp[i]).name, ii);
			DDefDouble4 (pre, sbuf, "m", &(((mp[i]).TLight[ii]).ds[1]), DVA_None);
			sprintf (sbuf, "%s.%d.ds.z", (mp[i]).name, ii);
			DDefDouble4 (pre, sbuf, "m", &(((mp[i]).TLight[ii]).ds[2]), DVA_None);
			
			sprintf (sbuf, "%s.%d.state", (mp[i]).name, ii);
			DDefInt (pre, sbuf, "", (int*) &(((mp[i]).TLight[ii]).state), DVA_None);
			
			sprintf (sbuf, "%s.%d.tRemain", (mp[i]).name, ii);
			DDefDouble4 (pre, sbuf, "s", &(((mp[i]).TLight[ii]).tRemain), DVA_None);
		}
		sprintf (sbuf, "%s.nLights", (mp[i]).name);
		DDefInt (pre, sbuf, "", &((mp[i]).nDtct), DVA_None);
	}
	
	DDefaultDelete(pre);
}

static void TLSensor_DeclQuants (void *MP) {
    struct tTLSensor *mp = MP;

    /* if MP == NULL, use static variables
    * else use dynamically allocated and actually active variables
    */
    if (mp == NULL) {
    /* Define here dict entries for non-dynamically allocated (static) variables. */

    } else {
    TLSensor_DeclQuants_dyn (mp, 0);
    }
}


/*
** TLSensor_New ()
** Read in s-v-course from file before TestRun start:
*/
static void * TLSensor_New (struct tInfos *Inf, const char *KindKey) {
	struct tTLSensor *mp = NULL;
	struct tTLight Light;
    int i;
	int ibuf[TSignSensorCount];
	int iTempMax[TSignSensorCount];
	int iTempMin[TSignSensorCount];
	char sbuf[99];
	char TempMaxbuf[99];
	char TempMinbuf[99];
	int routeid;
	tRoadTrfSignList trafficSignList;
	
	if (TSignSensorCount > 0) {

		/* For CarMaker 6.*:
		 * routeid = Env.Route.Id;
		 * For CarMaker 7.* and 8.*:
		 * routeid = RoadRouteObjIdGetRouteId(Env.Road, Env.Route.ObjId);
		 */
		routeid = RoadRouteObjIdGetRouteId(Env.Road, Env.Route.ObjId);

		if(RoadGetAllTrfSigns(Env.Road, routeid, &trafficSignList)== 0 ){
			LogWarnF(EC_Init, "The TLSensor does not detect any traffic lights as there are no traffic signs present and hence the traffic sign sensor does not function.\n");
			LogWarnF(EC_Init, "Insert a traffic sign in the scenario to enable the traffic light sensor.\n");
		}
	}

	/* Initialize sensor count */
	TLSensorCount = 0;
	/* Look for entries in Vehicle > Additional Parameters */
	for (i=0; i < TSignSensorCount; i++) {
		sprintf(sbuf, "TLSensor.%s.TrackLimit", (TSignSensor[i]).BS.Name);
		sprintf(TempMaxbuf, "TLSensor.%s.DtctAngle_z_Max", (TSignSensor[i]).BS.Name);
		sprintf(TempMinbuf, "TLSensor.%s.DtctAngle_z_Min", (TSignSensor[i]).BS.Name);
		ibuf[i] = iGetIntOpt(SimCore.Vhcl.Inf, sbuf, 0);
		iTempMax[i] = iGetDblOpt (SimCore.Vhcl.Inf, TempMaxbuf, 0);
		iTempMin[i] = iGetDblOpt (SimCore.Vhcl.Inf, TempMinbuf, 0);
		if (ibuf[i]>0) {
			TLSensorCount++;
		}
	}
	if (TLSensorCount > 0) { /* only allocate memory, if sensors defined */
		mp = (struct tTLSensor *)calloc(TLSensorCount, sizeof(*mp));
		
		if (TLSensor_NewObject(&Light) < 0)
			LogErrStr(EC_Init, "[TLSensor] Failed to initialize dummy TLight-Object!");


		int ii=0;
		for (i=0; i < TSignSensorCount; i++) {
			if (ibuf[i] > 0) {

				sprintf((mp[ii]).name, "%s", (TSignSensor[i]).BS.Name);
				(mp[ii]).TSSid = i;
				(mp[ii]).trackLimit = ibuf[i];
				(mp[ii]).DtctAngle_z_Max = iTempMax[i];
				(mp[ii]).DtctAngle_z_Min = iTempMin[i];
				if (0)
					(mp[ii]).TLight = (struct tTLight *)calloc(TrfLight.nObjs, sizeof(tTLight));
				else
					(mp[ii]).TLight = (struct tTLight *)calloc((mp[ii]).trackLimit, sizeof(tTLight));
				ii++;
			}

		}

	} else {
			LogWarnF(EC_Init, "TLSensor Cannot be initialized as Track limit of atleast one of the sensors is not defined. \n");
			mp = (struct tTLSensor *)calloc(1, sizeof(*mp));
	}
    
    return mp;
	
    return NULL;
}

/*
** TLSensor_Calc ()
*/
static int TLSensor_Calc (void *MP, double dt) {
    struct tTLSensor *mp = MP;
	struct tTLight TempObj;
	
    /* Execute only during simulation and if TLSensor is active */
    if ((SimCore.State == SCState_Simulate) && (TLSensorCount > 0)) {
		int tls, i, r;
		int UpdTime; 			// Update cycle time of sensor [ms]
								// deducted from TSSensor -> UpdRate [Hz]
		double T[3];			// Temporary vector
		double S[3]; 			// Vector Fr0 - Sensor
		double TL[3]; 			// Vector Fr0 - Traffic Light
		double X0[3]; 			// Vector Sensor - Traffic Light in Fr0
		double XS[3]; 			// Vector Sensor - Traffic Light in FrS
		double U0[3]; 			// Orientation vector Traffic Light in Fr0
		double US[3]; 			// Orientation vector Traffic Light in FrS
		double a; 				// Azimuth of r-x in rad, sin and cos
		double t; 				// Elevation of r-x in rad, sin and cos
		double rz;				// Orientation angle of Traffic Light around z
		double ry;				// Orientation angle of Traffic Light around y
		double rx;				// Orientation angle of Traffic Light around x
		int v;					// valid flag
		
		VEC_Assign(X0, Null3x1);
		/* Cycle through sensors */
		for (tls=0; tls < TLSensorCount; tls++) {
			
			UpdTime = (int) M_ROUND(1000 / (TSignSensor[(mp[tls]).TSSid]).UpdRate);
			if ((SimCore.CycleNo % UpdTime) == 0)
			{
				(mp[tls]).nDtct = 0;
				TLSensor_NewObject (&TempObj);
				
				/* Reset all detection list */
				for (i=0; i < (mp[tls]).trackLimit; i++) {
					TLSensor_CopyObject (&((mp[tls]).TLight[i]), TempObj);
				}
				/* Cycle through all traffic lights, as routes might not be used */
				for (i=0; i < TrfLight.nObjs; i++) {
					if ((v = TLSensor_IsValidType(TrfLight.Objs[i].TLList[0])) < 0) {
						return TLSensor_ERROR;
                    }
					else if (!v)
						continue;		/* no further calculation for invalid types */
					
					// Read the Sensor position in FR0
					VEC_Assign(S, (TSignSensor[(mp[tls]).TSSid]).BS_Pos_0);

					// Read the Traffic Light Position in FR0
					VEC_Assign(TL, (TrfLight.Objs[i]).TLList[0].Pos_0);

					// Vector subtraction (X0= TL-S) to get the position vector of TL from S
					VEC_Sub(X0, TL, S);

					// Position Vector of Traffic Light in FRS
					VEC_MatTVec(XS, (TSignSensor[(mp[tls]).TSSid]).Tr2Fr0, X0);

					// Azimuth of the detected Traffic Light
					VEC_AssignByComp(T, XS[0], XS[1], 0);
					//a = asin(XS[1] / VEC_Norm(T));
					a = atan(XS[1]/XS[0]);

					// Elevation of the detected Traffic Light
					VEC_AssignByComp(T, XS[0], XS[1], XS[2]);
					//VEC_AssignByComp(T, XS[0], 0, XS[2]);
					t = asin(XS[2] / VEC_Norm(T));
					
					// Identify the quadrant in which the Traffic Light is detected; if 
					// the Traffic Light is detected in III or IV quadrant, it is filtered 
					// out in further steps
					if(XS[0]<0 && XS[1]>0){
						a = a + 180 * deg2rad;
					}else if(XS[0]<0 && XS[1]<0){
						a = a - 180 * deg2rad;
					}

					// Write the angle of the TL in U[0] = [0,0,rotZ wrt FRO]
					TLSensor_GetTLOrientation (U0, (TrfLight.Objs[i]));

					// The angle is converted to FRS by Vector subtraction from angle made 
					// by the Sensor on FR1 and the angle made by FR1 with Fr0
					US[0] = U0[0] - TSignSensor[(mp[tls]).TSSid].rot_zyx[0]*rad2deg - Vehicle.Fr1A.r_zyx[0]*rad2deg;
					US[1] = U0[1] - TSignSensor[(mp[tls]).TSSid].rot_zyx[1]*rad2deg - Vehicle.Fr1A.r_zyx[1]*rad2deg;
					US[2] = U0[2] - TSignSensor[(mp[tls]).TSSid].rot_zyx[2]*rad2deg - Vehicle.Fr1A.r_zyx[2]*rad2deg;

					//Angle is rotated by 180 degrees so that the Light facing towards the 
					// sensor is 0 degree and written in radians
					rz = (fmod(US[2]+180,360))*deg2rad;

					// If the angle obtained is negative, it is made positive by adding 
					// 360deg
					if(rz<0){
						rz = rz + (360*deg2rad);
					}
					// After making it positive, if the angle is more than 180 degrees, 
					// the angle is subtracted by 360 so that the angle is obtained in 
					// negative but in clockwise direction
					if (rz>180*deg2rad){
						rz = rz - (360*deg2rad);
					}

					ry =(fmod(US[1],360))*deg2rad;
					if (0) {
						ry = 0;
					}

					rx = (fmod(US[0],360))*deg2rad;
					
					/* fill in the temp object */
					TempObj.id = (TrfLight.Objs[i]).TLList[0].ObjId;
					TempObj.state = (TrfLight.Objs[i]).State;
					TempObj.tRemain = (TrfLight.Objs[i]).tRemain;
					VEC_Assign(TempObj.ds, XS);
					TempObj.ds_p = VEC_Norm(XS);
					TempObj.alpha = a;
					TempObj.theta = t;
					VEC_AssignByComp(TempObj.orientation, rx, ry, rz);

					// Until now all the TL in the scenario are detected
					// In the following step, it is limited by 
					// - range
					// - field of view (alpha, theta)
					// - angle of detection
					/*if ((TempObj.ds_p < (TSignSensor[(mp[tls]).TSSid]).range)
						&& (fabs(TempObj.alpha) < (TSignSensor[(mp[tls]).TSSid]).alpha)
						&& (fabs(TempObj.theta) < (TSignSensor[(mp[tls]).TSSid]).theta)
						&& (rz<((mp[tls]).DtctAngle_z_Max*deg2rad))
						&& (rz>((mp[tls]).DtctAngle_z_Min*deg2rad))
						) {*/
							r = TLSensor_AddObject (&(mp[tls]), TempObj);
					//}
					
				}
				
			}
			
		}
		
    }
	
    return TLSensor_SUCCESS;
}


/*
** TLSensor_Delete ()
** Clean up and reset
*/
static void TLSensor_Delete (void *MP) {
    struct tTLSensor *mp = MP;
	
	// if (*There is any sensors*)
    if (TLSensorCount > 0) {

		free (mp);
		mp = NULL;
		
	}
    
}

int 
TLSensor_Register (void)
{
    tModelClassDescr m;

    memset(&m, 0, sizeof(m));
    m.Generic.VersionId =  ThisVersionId;
    m.Generic.New =        TLSensor_New;
    m.Generic.Calc =       TLSensor_Calc;
    m.Generic.Delete =     TLSensor_Delete;
    m.Generic.DeclQuants = TLSensor_DeclQuants;

    return Model_Register(ModelClass_Generic, ThisModelKind, &m);
}

int 
TLSensor_NewObject (struct tTLight *NewObj)
{
	NewObj->id = 0;
	VEC_Assign(NewObj->ds, Null3x1);
	NewObj->ds_p = 0;
	NewObj->alpha = 0;
	NewObj->theta = 0;
	NewObj->state = State_Off;
	NewObj->tRemain = 0;
	VEC_Assign(NewObj->orientation, Null3x1);

	return TLSensor_SUCCESS;
};

int 
TLSensor_AddObject (struct tTLSensor *pSensor, struct tTLight NewObj)
{
	int i;
	
	for (i=0; i < pSensor->trackLimit; i++) {
		if ((pSensor->TLight[i]).id > 0) {
			// List position is not empty
			if (NewObj.ds_p < (pSensor->TLight[i]).ds_p) {
				// New object is closer to sensor, so move list
				TLSensor_MoveInList (pSensor, i, i+1);
				// and add new entry at this position
				TLSensor_CopyObject (&(pSensor->TLight[i]), NewObj);
				break;
			}
		} else {
			// List position is empty
			TLSensor_CopyObject (&(pSensor->TLight[i]), NewObj);
			pSensor->nDtct++;
			break;
		}
	}
	if (i == pSensor->trackLimit) {
		return TLSensor_OUTOFLIST;
	}

//	return TLSensor_SUCCESS;
	return i;
};

int 
TLSensor_CopyObject (struct tTLight *to, struct tTLight from)
{
	to->id = from.id;
	VEC_Assign(to->ds, from.ds);
	to->ds_p = from.ds_p;
	to->alpha = from.alpha;
	to->theta = from.theta;
	to->state = from.state;
	to->tRemain = from.tRemain;
	VEC_Assign(to->orientation, from.orientation);

	return TLSensor_SUCCESS;
};

int 
TLSensor_MoveInList (struct tTLSensor *pSensor, int from, int to)
{
	if (to == pSensor->trackLimit) {
		// Entry moves out of list boundary
		// so stop movement to stay in memory range
		return TLSensor_OUTOFLIST;
	}
	if ((pSensor->TLight[to]).id > 0){
		// List position is occupied, move list further
		TLSensor_MoveInList (pSensor, to, (to+1));
		// then place moved entry
		TLSensor_CopyObject (&(pSensor->TLight[to]), (pSensor->TLight[from]));
	} else {
		// End of current list reached (within boundary), add entry
		TLSensor_CopyObject (&(pSensor->TLight[to]), (pSensor->TLight[from]));
		pSensor->nDtct++;
	}

	return TLSensor_SUCCESS;
};

/* *** excerpt from <roadbuilder.h> for reference ***
typedef struct tRoadTLOut {
	double xyz[3];
	double angle;
	double roadAngle;
	int type;
	int valid;
} tRoadTLOut;
*/ 

int 
TLSensor_GetTLOrientation (double *orientation, struct tTrfLightObj TrfLight)
{
	tRoadTLOut r;
	
	if (RoadGetTrfLightObj(Env.Road, (tObjId) TrfLight.ObjId, &r) < 0)
		return TLSensor_ERROR;
	else {
		orientation[0] = 0;
		orientation[1] = 0;
		orientation[2] = r.angle;
	}

	
	return TLSensor_SUCCESS;
};

int
TLSensor_IsValidType (struct tTrfLightObj_TL TrfLight)
{
	tRoadTLOut 	r;
	int 		v = 0;
	
	/* (R)TLType < (R)TLT_Ped_R && (R)TLType != (R)TLT_100_RYG: 
	 * Pedestrian lights (last in list) or small RYG light (e.g. bycicle lanes) do not 
	 * return valid. */
	
	if (RoadGetTrfLightObj(Env.Road, (tObjId) TrfLight.ObjId, &r) < 0) {
		return TLSensor_ERROR;
    }
	else {
		/* For CarMaker 6.* and 7.*:
		 * ((r.type < TLT_Ped_R) && r.type!=TLT_100_RYG)
		 * For CarMaker 8.*:
		 * ((r.type < RTLT_Ped_R) && r.type!=RTLT_100_RYG)
		 */
		if ((r.type < RTLT_Ped_R) && r.type!=RTLT_100_RYG)
			v = 1;
	}
	
	return v;
	
};

