/*
******************************************************************************
**  CarMaker - Version 6.0.3
**  Vehicle Dynamics Simulation Toolkit
**
**  Copyright (C)   IPG Automotive GmbH
**                  Bannwaldallee 60             Phone  +49.721.98520.0
**                  76185 Karlsruhe              Fax    +49.721.98520.99
**                  Germany                      WWW    www.ipg-automotive.com
**
******************************************************************************
**
**  This example is not part of the CarMaker product. As such, it does not
**  receive official support or updates.
**
******************************************************************************
*/

#ifndef _MYMODELS_H__
#define _MYMODELS_H__

#ifdef __cplusplus
extern "C" {
#endif

#define SENS_SIZE_NAME 16
#define SENS_MAX_KEY 128

typedef enum {
	TLSensor_OUTOFLIST = 	-2,
	TLSensor_ERROR = 		-1,
	TLSensor_SUCCESS = 		0
} tTLSensorResult;

typedef struct tTLight {
	int 	id; 				/* global object ID */
	double 	ds[3];				/* Cartesian coordinates in sensor frame */
	double 	ds_p; 				/* polar distance from sensor view */
	double 	alpha; 				/* azimuth angle in sensor frame */
	double 	theta; 				/* elevation angle in sensor frame */
	tTLState 	state; 				/* current traffic light state */
	double 	tRemain; 			/* remaining time in current state */
	double 	orientation[3]; 	/* orientation of traffic light (in sensor frame) * 
								 * vector of polar coordinates (r, alpha, theta)  */
} tTLight;


typedef struct tTLSensor {
	char 			name[SENS_SIZE_NAME]; 	/* name string of traffic light */
	int 			TSSid; 					/* ID of traffic sign sensor */
	int 			nDtct; 					/* number of detected traffic lights */
	int 			trackLimit; 			/* maximum number of traffic lights being 
											 * monitored */
	double			DtctAngle_z_Max;		/* maximum angle of the TL in FRS until it is 
											 * detected */
	double			DtctAngle_z_Min;		/* minimum angle of the TL in FRS until it is 
											 * detected */
    tTLight 		*TLight; 				/* structure of traffic light data */
} tTLSensor;

extern int 			TLSensorCount;

int TLSensor_NewObject (struct tTLight *NewObj);
int TLSensor_AddObject (struct tTLSensor *pSensor, struct tTLight NewObj);
int TLSensor_CopyObject (struct tTLight *to, struct tTLight from);
int TLSensor_MoveInList (struct tTLSensor *pSensor, int from, int to);
int TLSensor_GetTLOrientation (double *orientation, struct tTrfLightObj TrfLight);
int TLSensor_IsValidType (struct tTrfLightObj_TL TrfLight);


int TLSensor_Register	(void);

#ifdef __cplusplus
}
#endif

#endif	/* #ifndef _MYMODELS_H__ */
